@@ -1,27 +1,16 @@
 /*
- *	$Id: setpci.c,v 1.4 1998/06/09 19:22:05 mj Exp $
+ *	The PCI Utilities -- Manipulate PCI Configuration Registers
  *
- *	Linux PCI Utilities -- Manipulate PCI Configuration Registers
- *
- *	Copyright (c) 1998 Martin Mares <mj@atrey.karlin.mff.cuni.cz>
+ *	Copyright (c) 1998--2006 Martin Mares <mj@ucw.cz>
  *
  *	Can be freely distributed and used under the terms of the GNU GPL.
  */
 
-#define _GNU_SOURCE
-
 #include <stdio.h>
 #include <string.h>
 #include <stdlib.h>
-#include <fcntl.h>
+#include <stdarg.h>
 #include <unistd.h>
-#include <errno.h>
-#include <asm/byteorder.h>
-
-#include <asm/unistd.h>
-#ifdef __GLIBC__
-#include <syscall-list.h>
-#endif
 
 #include "pciutils.h"
 
@@ -29,206 +18,136 @@ static int force;			/* Don't complain if no devices match */
 static int verbose;			/* Verbosity level */
 static int demo_mode;			/* Only show */
 
-struct device {
-  struct device *next;
-  byte bus, devfn, mark;
-  word vendid, devid;
-  int fd, need_write;
-};
+const char program_name[] = "setpci";
 
-static struct device *first_dev;
+static struct pci_access *pacc;
+
+struct value {
+  unsigned int value;
+  unsigned int mask;
+};
 
 struct op {
   struct op *next;
-  struct device **dev_vector;
+  struct pci_dev **dev_vector;
   unsigned int addr;
   unsigned int width;			/* Byte width of the access */
   int num_values;			/* Number of values to write; <0=read */
-  unsigned int values[0];
+  struct value values[0];
 };
 
 static struct op *first_op, **last_op = &first_op;
+static unsigned int max_values[] = { 0, 0xff, 0xffff, 0, 0xffffffff };
 
-void *
-xmalloc(unsigned int howmuch)
-{
-  void *p = malloc(howmuch);
-  if (!p)
-    {
-      fprintf(stderr, "setpci: Unable to allocate %d bytes of memory\n", howmuch);
-      exit(1);
-    }
-  return p;
-}
-
-/*
- * As libc doesn't support pread/pwrite yet, we have to call them directly
- * or use lseek/read/write instead.
- */
-#ifdef __GLIBC__
-#ifndef SYS_pread
-#define SYS_pread __NR_pread
-#endif
-static int
-pread(unsigned int fd, void *buf, size_t size, loff_t where)
-{
-  return syscall(SYS_pread, fd, buf, size, where);
-}
-
-#ifndef SYS_pwrite
-#define SYS_pwrite __NR_pwrite
-#endif
-static int
-pwrite(unsigned int fd, void *buf, size_t size, loff_t where)
-{
-  return syscall(SYS_pwrite, fd, buf, size, where);
-}
-#else
-static _syscall4(int, pread, unsigned int, fd, void *, buf, size_t, size, loff_t, where);
-static _syscall4(int, pwrite, unsigned int, fd, void *, buf, size_t, size, loff_t, where);
-#endif
-
-static void
-scan_devices(void)
-{
-  struct device **last = &first_dev;
-  byte line[256];
-  FILE *f;
-
-  if (!(f = fopen(PROC_BUS_PCI "/devices", "r")))
-    {
-      perror(PROC_BUS_PCI "/devices");
-      exit(1);
-    }
-  while (fgets(line, sizeof(line), f))
-    {
-      struct device *d = xmalloc(sizeof(struct device));
-      unsigned int dfn, vend;
-
-      sscanf(line, "%x %x", &dfn, &vend);
-      d->bus = dfn >> 8U;
-      d->devfn = dfn & 0xff;
-      d->vendid = vend >> 16U;
-      d->devid = vend & 0xffff;
-      d->fd = -1;
-      *last = d;
-      last = &d->next;
-    }
-  fclose(f);
-  *last = NULL;
-}
-
-static struct device **
+static struct pci_dev **
 select_devices(struct pci_filter *filt)
 {
-  struct device *z, **a, **b;
+  struct pci_dev *z, **a, **b;
   int cnt = 1;
 
-  for(z=first_dev; z; z=z->next)
-    if (z->mark = filter_match(filt, z->bus, z->devfn, z->vendid, z->devid))
+  for(z=pacc->devices; z; z=z->next)
+    if (pci_filter_match(filt, z))
       cnt++;
   a = b = xmalloc(sizeof(struct device *) * cnt);
-  for(z=first_dev; z; z=z->next)
-    if (z->mark)
+  for(z=pacc->devices; z; z=z->next)
+    if (pci_filter_match(filt, z))
       *a++ = z;
   *a = NULL;
   return b;
 }
 
 static void
-exec_op(struct op *op, struct device *dev)
+exec_op(struct op *op, struct pci_dev *dev)
 {
-  char *mm[] = { NULL, "%02x", "%04x", NULL, "%08x" };
-  char *m = mm[op->width];
-  unsigned int x;
-  int i;
-  __u32 x32;
-  __u16 x16;
-  __u8 x8;
-
-  if (!demo_mode && dev->fd < 0)
-    {
-      char name[64];
-      sprintf(name, PROC_BUS_PCI "/%02x/%02x.%x", dev->bus, PCI_SLOT(dev->devfn), PCI_FUNC(dev->devfn));
-      if ((dev->fd = open(name, dev->need_write ? O_RDWR : O_RDONLY)) < 0)
-	{
-	  perror(name);
-	  exit(1);
-	}
-    }
+  char *formats[] = { NULL, "%02x", "%04x", NULL, "%08x" };
+  char *mask_formats[] = { NULL, "%02x->(%02x:%02x)->%02x", "%04x->(%04x:%04x)->%04x", NULL, "%08x->(%08x:%08x)->%08x" };
+  unsigned int x, y;
+  int i, addr;
+  int width = op->width;
 
   if (verbose)
-    printf("%02x:%02x.%x:%02x", dev->bus, PCI_SLOT(dev->devfn), PCI_FUNC(dev->devfn), op->addr);
+    printf("%02x:%02x.%x:%02x", dev->bus, dev->dev, dev->func, op->addr);
+  addr = op->addr;
   if (op->num_values >= 0)
-    for(i=0; i<op->num_values; i++)
-      {
-	if (verbose)
-	  {
-	    putchar(' ');
-	    printf(m, op->values[i]);
-	  }
-	if (demo_mode)
-	  continue;
-	switch (op->width)
-	  {
-	  case 1:
-	    x8 = op->values[i];
-	    i = pwrite(dev->fd, &x8, 1, op->addr);
-	    break;
-	  case 2:
-	    x16 = __cpu_to_le16(op->values[i]);
-	    i = pwrite(dev->fd, &x16, 2, op->addr);
-	    break;
-	  default:
-	    x32 = __cpu_to_le32(op->values[i]);
-	    i = pwrite(dev->fd, &x32, 4, op->addr);
-	    break;
-	  }
-	if (i != (int) op->width)
-	  {
-	    fprintf(stderr, "Error writing to %02x:%02x.%d: %m\n", dev->bus, PCI_SLOT(dev->devfn), PCI_FUNC(dev->devfn));
-	    exit(1);
-	  }
-      }
+    {
+      for(i=0; i<op->num_values; i++)
+	{
+	  if ((op->values[i].mask & max_values[width]) == max_values[width])
+	    {
+	      x = op->values[i].value;
+	      if (verbose)
+		{
+		  putchar(' ');
+		  printf(formats[width], op->values[i].value);
+		}
+	    }
+	  else
+	    {
+	      switch (width)
+		{
+		case 1:
+		  y = pci_read_byte(dev, addr);
+		  break;
+		case 2:
+		  y = pci_read_word(dev, addr);
+		  break;
+		default:
+		  y = pci_read_long(dev, addr);
+		  break;
+		}
+	      x = (y & ~op->values[i].mask) | op->values[i].value;
+	      if (verbose)
+		{
+		  putchar(' ');
+		  printf(mask_formats[width], y, op->values[i].value, op->values[i].mask, x);
+		}
+	    }
+	  if (!demo_mode)
+	    {
+	      switch (width)
+		{
+		case 1:
+		  pci_write_byte(dev, addr, x);
+		  break;
+		case 2:
+		  pci_write_word(dev, addr, x);
+		  break;
+		default:
+		  pci_write_long(dev, addr, x);
+		  break;
+		}
+	    }
+	  addr += width;
+	}
+      if (verbose)
+	putchar('\n');
+    }
   else
     {
       if (verbose)
 	printf(" = ");
-      if (!demo_mode)
+      switch (width)
 	{
-	  switch (op->width)
-	    {
-	    case 1:
-	      i = pread(dev->fd, &x8, 1, op->addr);
-	      x = x8;
-	      break;
-	    case 2:
-	      i = pread(dev->fd, &x16, 2, op->addr);
-	      x = __le16_to_cpu(x16);
-	      break;
-	    default:
-	      i = pread(dev->fd, &x32, 4, op->addr);
-	      x = __le32_to_cpu(x32);
-	      break;
-	    }
-	  if (i != (int) op->width)
-	    {
-	      fprintf(stderr, "Error reading from %02x:%02x.%d: %m\n", dev->bus, PCI_SLOT(dev->devfn), PCI_FUNC(dev->devfn));
-	      exit(1);
-	    }
-	  printf(m, x);
+	case 1:
+	  x = pci_read_byte(dev, addr);
+	  break;
+	case 2:
+	  x = pci_read_word(dev, addr);
+	  break;
+	default:
+	  x = pci_read_long(dev, addr);
+	  break;
 	}
-      else
-	putchar('?');
+      printf(formats[width], x);
+      putchar('\n');
     }
-  putchar('\n');
 }
 
 static void
 execute(struct op *op)
 {
-  struct device **vec = NULL;
-  struct device **pdev, *dev;
+  struct pci_dev **vec = NULL;
+  struct pci_dev **pdev, *dev;
   struct op *oops;
 
   while (op)
@@ -245,27 +164,21 @@ execute(struct op *op)
 static void
 scan_ops(struct op *op)
 {
-  struct device **pdev, *dev;
-
   while (op)
     {
       if (op->num_values >= 0)
-	{
-	  pdev = op->dev_vector;
-	  while (dev = *pdev++)
-	    dev->need_write = 1;
-	}
+	pacc->writeable = 1;
       op = op->next;
     }
 }
 
 struct reg_name {
-  int offset;
-  int width;
-  char *name;
+  unsigned int offset;
+  unsigned int width;
+  const char *name;
 };
 
-static struct reg_name pci_reg_names[] = {
+static const struct reg_name pci_reg_names[] = {
   { 0x00, 2, "VENDOR_ID", },
   { 0x02, 2, "DEVICE_ID", },
   { 0x04, 2, "COMMAND", },
@@ -333,22 +246,30 @@ static struct reg_name pci_reg_names[] = {
   { 0x00, 0, NULL }
 };
 
-static void usage(void) __attribute__((noreturn));
-
-static void
-usage(void)
+static void NONRET
+usage(char *msg, ...)
 {
+  va_list args;
+  va_start(args, msg);
+  if (msg)
+    {
+      fprintf(stderr, "setpci: ");
+      vfprintf(stderr, msg, args);
+      fprintf(stderr, "\n\n");
+    }
   fprintf(stderr,
-"Usage: setpci [-fvD] (<device>+ <reg>[=<values>]*)*\n\
+"Usage: setpci [<options>] (<device>+ <reg>[=<values>]*)*\n\
 -f\t\tDon't complain if there's nothing to do\n\
 -v\t\tBe verbose\n\
--D\t\tList changes, don't commit them\n\
-<device>:\t-s [[<bus>]:][<slot>][.[<func>]]\n\
-\t|\t-d [<vendor>]:[<device>]\n\
-<reg>:\t\t<number>[.(B|W|L)]\n\
-     |\t\t<name>\n\
-<values>:\t<value>[,<value>...]\n\
-");
+-D\t\tList changes, don't commit them\n"
+GENERIC_HELP
+"<device>:\t-s [[[<domain>]:][<bus>]:][<slot>][.[<func>]]\n"
+"\t|\t-d [<vendor>]:[<device>]\n"
+"<reg>:\t\t<number>[.(B|W|L)]\n"
+"     |\t\t<name>\n"
+"<values>:\t<value>[,<value>...]\n"
+"<value>:\t<hex>\n"
+"       |\t<hex>:<mask>\n");
   exit(1);
 }
 
@@ -357,7 +278,8 @@ main(int argc, char **argv)
 {
   enum { STATE_INIT, STATE_GOT_FILTER, STATE_GOT_OP } state = STATE_INIT;
   struct pci_filter filter;
-  struct device **selected_devices = NULL;
+  struct pci_dev **selected_devices = NULL;
+  char *opts = GENERIC_OPTIONS ;
 
   if (argc == 2 && !strcmp(argv[1], "--version"))
     {
@@ -366,10 +288,15 @@ main(int argc, char **argv)
     }
   argc--;
   argv++;
+
+  pacc = pci_alloc();
+  pacc->error = die;
+
   while (argc && argv[0][0] == '-')
     {
       char *c = argv[0]+1;
       char *d = c;
+      char *e;
       while (*c)
 	switch (*c)
 	  {
@@ -388,16 +315,42 @@ main(int argc, char **argv)
 	  case 0:
 	    break;
 	  default:
-	    if (c != d)
-	      usage();
-	    goto next;
+	    if (e = strchr(opts, *c))
+	      {
+		char *arg;
+		c++;
+		if (e[1] == ':')
+		  {
+		    if (*c)
+		      arg = c;
+		    else if (argc > 1)
+		      {
+			arg = argv[1];
+			argc--; argv++;
+		      }
+		    else
+		      usage(NULL);
+		    c = "";
+		  }
+		else
+		  arg = NULL;
+		if (!parse_generic_option(*e, pacc, arg))
+		  usage(NULL);
+	      }
+	    else
+	      {
+		if (c != d)
+		  usage(NULL);
+		goto next;
+	      }
 	  }
       argc--;
       argv++;
     }
 next:
 
-  scan_devices();
+  pci_init(pacc);
+  pci_scan_bus(pacc);
 
   while (argc)
     {
@@ -405,49 +358,44 @@ next:
       char *d, *e, *f;
       int n, i;
       struct op *op;
-      unsigned long ll, lim;
+      unsigned long ll;
+      unsigned int lim;
 
       if (*c == '-')
 	{
 	  if (!c[1] || !strchr("sd", c[1]))
-	    usage();
+	    usage(NULL);
 	  if (c[2])
 	    d = (c[2] == '=') ? c+3 : c+2;
-	  else if (argc)
+	  else if (argc > 1)
 	    {
 	      argc--;
 	      argv++;
 	      d = argv[0];
 	    }
 	  else
-	    usage();
+	    usage(NULL);
 	  if (state != STATE_GOT_FILTER)
 	    {
-	      filter_init(&filter);
+	      pci_filter_init(pacc, &filter);
 	      state = STATE_GOT_FILTER;
 	    }
 	  switch (c[1])
 	    {
 	    case 's':
-	      if (d = filter_parse_slot(&filter, d))
-		{
-		  fprintf(stderr, "setpci: -s: %s\n", d);
-		  return 1;
-		}
+	      if (d = pci_filter_parse_slot(&filter, d))
+		die("-s: %s", d);
 	      break;
 	    case 'd':
-	      if (d = filter_parse_id(&filter, d))
-		{
-		  fprintf(stderr, "setpci: -d: %s\n", d);
-		  return 1;
-		}
+	      if (d = pci_filter_parse_id(&filter, d))
+		die("-d: %s", d);
 	      break;
 	    default:
-	      usage();
+	      usage(NULL);
 	    }
 	}
       else if (state == STATE_INIT)
-	usage();
+	usage(NULL);
       else
 	{
 	  if (state == STATE_GOT_FILTER)
@@ -455,16 +403,17 @@ next:
 	  if (!selected_devices[0] && !force)
 	    fprintf(stderr, "setpci: Warning: No devices selected for `%s'.\n", c);
 	  state = STATE_GOT_OP;
+	  /* look for setting of values and count how many */
 	  d = strchr(c, '=');
 	  if (d)
 	    {
 	      *d++ = 0;
 	      if (!*d)
-		usage();
+		usage("Missing value");
 	      for(e=d, n=1; *e; e++)
 		if (*e == ',')
 		  n++;
-	      op = xmalloc(sizeof(struct op) + n*sizeof(unsigned int));
+	      op = xmalloc(sizeof(struct op) + n*sizeof(struct value));
 	    }
 	  else
 	    {
@@ -478,7 +427,7 @@ next:
 	    {
 	      *e++ = 0;
 	      if (e[1])
-		usage();
+		usage("Missing width");
 	      switch (*e & 0xdf)
 		{
 		case 'B':
@@ -488,7 +437,7 @@ next:
 		case 'L':
 		  op->width = 4; break;
 		default:
-		  usage();
+		  usage("Invalid width \"%c\"", *e);
 		}
 	    }
 	  else
@@ -496,37 +445,48 @@ next:
 	  ll = strtol(c, &f, 16);
 	  if (f && *f)
 	    {
-	      struct reg_name *r;
+	      const struct reg_name *r;
 	      for(r = pci_reg_names; r->name; r++)
 		if (!strcasecmp(r->name, c))
 		  break;
-	      if (!r->name || e)
-		usage();
+	      if (!r->name)
+		usage("Unknown register \"%s\"", c);
+	      if (e && op->width != r->width)
+		usage("Explicit width doesn't correspond with the named register \"%s\"", c);
 	      ll = r->offset;
 	      op->width = r->width;
 	    }
-	  if (ll > 0x100 || ll + op->width*((n < 0) ? 1 : n) > 0x100)
-	    {
-	      fprintf(stderr, "setpci: Register number out of range!\n");
-	      return 1;
-	    }
+	  if (ll > 0x1000 || ll + op->width*((n < 0) ? 1 : n) > 0x1000)
+	    die("Register number out of range!");
 	  if (ll & (op->width - 1))
-	    {
-	      fprintf(stderr, "setpci: Unaligned register address!\n");
-	      return 1;
-	    }
+	    die("Unaligned register address!");
 	  op->addr = ll;
+	  /* read in all the values to be set */
 	  for(i=0; i<n; i++)
 	    {
 	      e = strchr(d, ',');
 	      if (e)
 		*e++ = 0;
 	      ll = strtoul(d, &f, 16);
-	      lim = (2 << ((op->width << 3) - 1)) - 1;
-	      if (f && *f ||
-		  (ll > lim && ll < ~0UL - lim))
-		usage();
-	      op->values[i] = ll;
+	      lim = max_values[op->width];
+	      if (f && *f && *f != ':')
+		usage("Invalid value \"%s\"", d);
+	      if (ll > lim && ll < ~0UL - lim)
+		usage("Value \"%s\" is out of range", d);
+	      op->values[i].value = ll;
+	      if (f && *f == ':')
+		{
+		  d = ++f;
+		  ll = strtoul(d, &f, 16);
+		  if (f && *f)
+		    usage("Invalid mask \"%s\"", d);
+		  if (ll > lim && ll < ~0UL - lim)
+		    usage("Mask \"%s\" is out of range", d);
+		  op->values[i].mask = ll;
+		  op->values[i].value &= ll;
+		}
+	      else
+		op->values[i].mask = ~0U;
 	      d = e;
 	    }
 	  *last_op = op;
@@ -537,7 +497,7 @@ next:
       argv++;
     }
   if (state == STATE_INIT)
-    usage();
+    usage("No operation specified");
 
   scan_ops(first_op);
   execute(first_op);
